﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
//
using System.Threading;	//Sleep()

class LcdLib
{
	/*
	ポート／ビットの対応
	[AVR-CDC]	[LCD]
		PB5	---	DB7
		PB4	---	DB6
		PB3	---	DB5
		PB2	---	DB4
		PB1	---	E
		PB0	---	RS
	*/
	
	//使用するポート（単一のポート）
	private const SFR LCD_PORT	= SFR.PORTB;
	private const SFR LCD_DDR	= SFR.DDRB;

	//ポートのビットの割り当て（4bitモードが前提）
	private const int
		LCD_RS	= (1 << RB.PB0),	//Register Select	H:data	L:command
		LCD_RW	= 0,				//Read/Write		H:read	L:write		※使用せず（ハード的にwrite固定）
		LCD_E	= (1 << RB.PB1),	//Enable Signal		H->L:strobe
		LCD_DATABITS_SHIFT	= 2,	//PB2-PB5	（LSB側から何ビットめがデータビットか。例：bit3-0なら0、bit7-4なら4）
		LCD_DATABITS	= (0x0F << LCD_DATABITS_SHIFT);
	
	//Instruction
	public const int
		LCD_CMD_CLEAR_DISPLAY	= 0x01,	//0b00000001
		LCD_CMD_RETURN_HOME		= 0x02,	//0b00000010
		LCD_CMD_ENTRY_MODE		= 0x04,	//0b00000100
		LCD_CMD_DISPLAY_MODE	= 0x08,	//0b00001000
		LCD_CMD_CURSOR_MODE		= 0x10,	//0b00010000
		LCD_CMD_FUNCTION_SET	= 0x20,	//0b00100000
		LCD_CMD_SET_CGRAM_ADRS	= 0x40,	//0b01000000
		LCD_CMD_SET_DDRAM_ADRS	= 0x80;	//0b10000000
	
	//Instructionのパラメータ
	//(LCD_CMD_AAA | LCD_SUBCMD_XXX1 | LCD_SUBCMD_XXX2 | …)の形で使う
	public const int
	//LCD_CMD_ENTRY_MODE
		LCD_SUBCMD_INCREMENT	= 0x02,	//0b00000010	//入力位置を進める（無指定だとDecrement）
		LCD_SUBCMD_WITH_SHIFT	= 0x01,	//0b00000001	//表示シフト動作を伴う（LCD_CMD_CURSOR_MODEも設定する）
	//LCD_CMD_DISPLAY_MODE
		LCD_SUBCMD_SET_DISPLAY	= 0x04,	//0b00000100	//画面表示する（無指定だと表示内容を保ったまま画面を消す）
		LCD_SUBCMD_SET_CURSOR	= 0x02,	//0b00000010	//カーソルを表示する（アンダーバーを表示する）
		LCD_SUBCMD_CURSOR_BLINK	= 0x01,	//0b00000001	//カーソルを点滅させる（文字に被さって黒マスが点滅する）
	//LCD_CMD_CURSOR_MODE
		LCD_SUBCMD_SHIFT_ON		= 0x08,	//0b00001000	//DDRAMの変更無しに表示をシフトさせる（無指定だとカーソルを移動させる）
		LCD_SUBCMD_RIGHT		= 0x04,	//0b00000100	//右へシフトさせる（無指定だと左へ）※上記、カーソル移動モード時は意味を持たない
	//LCD_CMD_FUNCTION_SET
		LCD_SUBCMD_DL8BIT		= 0x10,	//0b00010000	//Data Length 8-bit（無指定だと4-bitモード）
		LCD_SUBCMD_2LINE		= 0x08,	//0b00001000	//表示で2行使用する（無指定だと1行のみ使用可能になる）
		LCD_SUBCMD_FONT5X10		= 0x04;	//0b00000100	//Font Type 5x10（無指定だと5x8）※SC1602/SD1602は物理的に非対応

	private CdcIO cdcio;
	private Encoding sjis;
	
	//コンストラクタ
	public LcdLib()
	{
		cdcio = null;
		sjis = Encoding.GetEncoding("shift_jis");
	}
	
	//CdcIOオブジェクトをセットする
	public void setCdcIO(CdcIO cdcio)
	{
		this.cdcio = cdcio;
	}
	
	private void _delay_ms(int ms)
	{
		Thread.Sleep(ms);
	}
	
	private void _delay_us(int us)
	{
		Thread.Sleep(1);
	}
	
	private void cmd4bit(int data)
	{
		//dataは4bitであることが前提
		int bits = LCD_E | (data << LCD_DATABITS_SHIFT);
		cdcio.Set(LCD_PORT, bits);
		_delay_ms(1);
		cdcio.SetXor(LCD_PORT, LCD_E);
		_delay_ms(1);
	}
	
	//コマンド実行
	public void cmd(int data)
	{
		cmd4bit((data >> 4) & 0x0F);	//上位4bit
		cmd4bit(data & 0x0F);			//下位4bit
	}

	//LCD初期化
	public void init()
	{
		cdcio.Set(LCD_DDR, LCD_E | LCD_RW | LCD_RS | LCD_DATABITS);
		
		_delay_ms(15);
		cmd4bit(0x03);		//0b0011
		_delay_ms(5);
		cmd4bit(0x03);		//0b0011
		_delay_us(100);
		cmd4bit(0x03);		//0b0011
		
		//4-bitモードへ移行
		cmd4bit(0x02);		//0b0010
		
		//----------bit:76543210
		cmd(0x28);	//0b00101000	//Function Set	bit:5(1), DL=4-bit(0), N=2lines(1), F=5x8dots(0), *(0), *(0)
		cmd(0x08);	//0b00001000	//Display Set	bit:3(1), D=Set Display:NO(0), C=Set Cursor:NO(0), B=Cursor Blink:NO(0)
		cmd(0x01);	//0b00000001	//Clear Display	bit:0(1)
		cmd(0x06);	//0b00000110	//Entry Mode	bit:2(1), I/D=Increment(1), S=With Display Shift:NO(0)
		cmd(0x0C);	//0b00001100	//Display Set	bit:3(1), D=Set Display:YES(1), C=Set Cursor:NO(0), B=Cursor Blink:NO(0)
	}

	//LCD画面消去
	public void cls()
	{
		cmd(LCD_CMD_CLEAR_DISPLAY);
	}

	//表示位置指定
	//引数	x:	桁 0,1,2,...
	//		y:	行 0,1 のみ（2行タイプのLCDモジュール専用）
	public void setPos(int x, int y)
	{
		if (x < 0 || y < 0 || 1 < y) return;
		cmd(LCD_CMD_SET_DDRAM_ADRS | (y * 0x40 + x));
	}
	
	//LCDで使用しているポートのビットをリセットする
	public void bitReset()
	{
		cdcio.Set(LCD_PORT, 0x00);
	}

	//1文字出力する
	public void putChar(char c)
	{
		int bits = LCD_RS | LCD_E | (((c >> 4) & 0x0F) << LCD_DATABITS_SHIFT);	//上位4bit
		cdcio.Set(LCD_PORT, bits);
		_delay_us(100);
		cdcio.SetXor(LCD_PORT, LCD_E);
		_delay_us(100);

		bits = LCD_RS | LCD_E | ((c & 0x0F) << LCD_DATABITS_SHIFT);		//下位4bit
		cdcio.Set(LCD_PORT, bits);
		_delay_us(100);
		cdcio.SetXor(LCD_PORT, LCD_E);
		_delay_us(100);
		cdcio.SetXor(LCD_PORT, LCD_RS);
		_delay_us(100);
	}

	//文字列を出力する
	public void putStr(string s)
	{
		if (s == null) return;
		byte[] bytes = sjis.GetBytes(s);
		for (int i = 0; i < bytes.Length; i++)
			putChar((char)bytes[i]);
		bytes = null;
		System.GC.Collect();
	}

	//数値を出力する
	public void putInt(int n)
	{
		//参考：WinAVRの場合 (0 <= uint16_t <= 65535) (-32768 <= int16_t < 32767)
		putStr(n.ToString());
	}

	//文字を定義する
	//引数	id:		0～7（キャラクタコード）
	//		dots:	5x8のドットパターンを格納した配列
	public void defChar(int id, byte[] dots)
	{
		if (dots.Length < 8) return;
		for (int i = 0; i < 8; i++)
		{
			cmd(LCD_CMD_SET_CGRAM_ADRS | (id << 3) | i);
			putChar((char)dots[i]);
		}
	}
}
